#!/usr/bin/python
# -*- coding: utf-8 -*-
##   text_utils.py 
##
##   Author: Saeed Rasooli <saeed.gnu@gmail.com>  (ilius)
##
##   This program is a free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 3, or (at your option)
##   any later version.
##
##   You can get a copy of GNU General Public License along this program
##   But you can always get it from http://www.gnu.org/licenses/gpl.txt
##
##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  the
##   GNU General Public License for more details.

import string, sys, os
##from  xml.etree.ElementTree import XML, tostring  ## used for xml2dict


## ascii spacial characters.
schAs=["\n", ",", ".", "[", "]", "(", ")", '-', '+', '=', '/', '\', ',':',';', "'", '"','`','_','!','?','*','@','#']

digitsFa=['۰','۱','۲','۳','۴','۵','۶','۷','۸','۹']

## persian spacial characters.
schFa=[
'\xd8\x9b', '\xd8\x9f','\xe2\x80\xa6','\xc2\xab','\xc2\xbb','\xd9\x80','\xd9\x94','\xd8\x8c' ,'\xe2\x80\x93','\xe2\x80\x9c','\xe2\x80\x9d','\xe2\x80\x8c']
commaFa='\xd8\x8c'

## other unicode spacial characters.
schUn=['\xee\x80\x8a','\xee\x80\x8c']

#for ch in schFa + schUn :
#  print ch,
sch = schAs+schFa+schUn+list(string.whitespace)+list(string.digits)+digitsFa

def myPrint(s):
  print s,
  sys.stdout.flush()

def textProgress(n=100, t=0.1):
  import time
  for i in range(n):
    myPrint('\b\b\b\b##%3d'%(i+1));time.sleep(t)
  myPrint('\b\b\b')

def printAsError(text='An error occured!', exit=False):
  (prev_stdout, sys.stdout) = (sys.stdout, sys.stderr)
  print text
  sys.stdout = prev_stdout
  if exit:
    sys.exit(1)



def timeHMS(seconds):
  (h,m,s)=time.gmtime(int(seconds))[3:6]
  if h==0:
    if m==0:
      return str(s).zfill(2)
    else:
      return str(m).zfill(2) + ':' + str(s).zfill(2)
  else:
      return str(h)          + ':' + str(m).zfill(2) + ':' +str(s).zfill(2)


def addDefaultOptions(opt, defOpt, escapeList=[None,'Unknown','unknown']):
  # Two varable opt(meaning options) and defOpt(meaning defaults options) have dict type.
  # this function sets options to defaults if they have not defined 
  # or have special values (in escapeList)
  # modifying opt variable an reurning nothing.
  optc = opt.copy()
  for item in defOpt.keys():
    if item in optc.keys():
      if not optc[item] in escapeList:
        continue
    optc[item] = defOpt[item]
  return optc

def mergeLists(lists):
  if not type(lists) in (list, tuple):
    raise TypeError, 'bad type given to mergeLists: %s'%myType(lists) 
  """
  for i in range(len(lists)):
    item = lists[i]
    if type(item)!=list:
      raise TypeError, 'argument give to mergeLists() at index %d is: \'%s\' ,bad type: \'%s\'' % (i, item, myType(item))
  """
  if len(lists)==0:
    return []
  elif len(lists)==1:
    if type(lists[0])==list:
      return lists[0][:]
    else:
      return lists[0]
  else:
    return lists[0]+mergeLists(lists[1:])

def findAll(st, sub):
  ind = []
  if type(sub)==type(' '):
    i = 0
    sbl = len(sub)
    while True:
      i = string.find(st, sub, i)
      if i==-1:
        break
      ind += [i]
      i += sbl
  elif type(sub) in [list,type(('',))]:
    for item in sub:
      ind += findAll(st, item)
    ind.sort()
  else:
    print 'Invailed second argument to function findAll!' 
    return []
  return ind


def sortby(lst, n, reverse=False):
    nlist = [(x[n], x) for x in lst]
    nlist.sort(None, None, reverse)
    return [val for (key, val) in nlist]


def sortby_inplace(lst, n, reverse=False):
    lst[:] = [(x[n], x) for x in lst]
    lst.sort(None, None, reverse)
    lst[:] = [val for (key, val) in lst]
    return


"""
def locate(lst, val):
 n=len(lst)
 if n==0:
  return
 if val < lst[0]:
  return -0.5
 if val == lst[0]:
  return 0
 if val == lst[-1]:
  return n-1
 if val > lst[-1]:
  return n-0.5
 si=0 # start index
 ei=n # end index
 while ei-si>1:
  mi=(ei+si)/2 # middle index
  if lst[mi] == val :
   return mi
  elif lst[mi] > val :
   ei=mi
   continue
  else:
   si=mi
   continue
 if ei-si==1:
  return si+0.5


def locate2(lst, val, ind=1):
 n=len(lst)
 if n==0:
  return
 if val<lst[0][ind]:
  return -0.5
 if val==lst[0][ind]:
  return 0
 if val==lst[-1][ind]:
  return n-1
 if val>lst[-1][ind]:
  return n-0.5
 si=0
 ei=n
 while ei-si>1:
  mi=(ei+si)/2
  if lst[mi][ind]==val:
   return mi
  elif lst[mi][ind]>val:
   ei=mi
   continue
  else:
   si=mi
   continue
 if ei-si==1:
  return si+0.5
"""


def checkOrder(lst):
 wrong=[]
 for i in range(len(lst)-1):
   if lst[i] == lst[i+1]:
     wrong += [i]
 return wrong

def removeRepeats(lst):
  ## gets a sorted list and removes any reapeated member. returns result.
  n=len(lst)
  if n==0:
    return []
  lstR=[lst[0]]
  for i in range(1,n):
    if lst[i] != lst[i-1]:
      lstR += [lst[i]]
  return lstR

def validWord(word): 
  if (word=="" or word==" " or word=="\n" ):
    return False
  else:
    return True 

def addWord(word, allWords):
    if len(allWords)==0: 
      return [word]
    i=locate(allWords,word)
    ii=int(i+1)
    if ii-i==0.5:
      allWords.insert(ii,word)
    return allWords

def findWords(st0, opt={}):
   # take all words of a text
   # and returns their indexes as a list.
   defOpt = {'minLen':3, 'noEn':True}
   opt = addDefaultOptions(opt, defOpt)
   st = st0[:]
   ind = []
   for ch in sch:
     st = st.replace(ch,' '*len(ch))
   if len(st)!=len(st0):
     print 'Error in function text_utlis.findWord. string length has been changed!';return []
   si = [-1]+findAll(st, ' ')+[len(st)] # separatior indexes
   for i in range(len(si)-1):
     word=st[si[i]+1:si[i+1]]
     if word.replace(' ','')=='':
       continue
     if 'word' in opt.keys():
       if word != opt['word']:
         continue  
     if len(word) < opt['minLen']:
       continue
     if opt['noEn']:
       en = False
       for c in word:
         if c in string.printable:
           en = True
       if en:
         continue
     ind.append( [ si[i]+1 , si[i+1] ] )
   return ind

def takeStrWords(st, opt={}):
   # take all words of a text
   # and returns them as a list of strings.
   defOpt={'minLen':3, 'noEn':True, 'sort':True, 'noRepeat':True}
   opt=addDefaultOptions(opt, defOpt)
   words = [ st[i:j] for [i,j] in findWords(st, opt) ]
    # 'sort' and 'noRepeat' options will not be used in findWords()
   if opt['sort']:
       words.sort()
   if opt['noRepeat']:
       words=removeRepeats(words)
   return words

def takeFileWords(filePath, opt={'minLen':3, 'sort':True, 'noRepeat':True}):
  try:
    fp = file(filePath,'r')
  except:
    print 'Can not open file',filePath
  return takeStrWords(fp.read(), opt)


#def removeTextTags(st1, removeBetween=False, tags=['i', 'I', 'b', 'B', 'a', 'A']):
def removeTextTags(st1, removeBetween, tags):
 st=st1[:]
 for tag in tags:
   i0=0
   while True:
     i1=st.find('<%s'%tag, i0)
     if i1==-1:
       break
     i2=st.find('>', i1+1)
     if i2==-1:
       i0=i1+1
       continue
     i3=st.find('</%s'%tag, i2+1)
     if i3==-1:
       st=st[:i1]+st[i2+1:]
       i0=i2+1
       continue
     i4=st.find('>',i3+1)
     if i4==-1:
       i0=i3+1
       continue
     if removeBetween:
       st = st[:i1] + st[i4+1:]
     else:
       st = st[:i1] + st[i2+1:i3] + st[i4+1:]
     i0=i4+1
   #print tag,'\t',i1,i2,i3,i4
 return st

def replaceInFileAs2Files(inF, fromF, toF, outName):
  lines1 = fromF.readlines()
  lines2 = toF.readlines()
  n = min(len(lines1), len(lines2))
  rpl = []
  for i in range(n):
    if lines1[i] != lines2[i]:
      rpl += [ [ lines1[i][:-1], lines2[i][:-1] ] ]
  del lines1, lines2
  text = inF.read()
  for item in rpl:
    text = text.replace(item[0], item[1])
  outF = file(outName)
  outF.write(text)
  outF.close()
  del text, outF

"""
def xml2dict(xmlText):
  from  xml.etree.ElementTree import XML, tostring
  xmlElems = XML(xmlText)
  for elem in xmlElems:
    elemText=tostring(elem)
    try:
      elem[0]
      elemElems=xml2dict()
    except:
      pass
"""

def relation(word, phrase, opt={}):
      defOpt={'sep':commaFa, 'matchWord':True}
      opt = addDefaultOptions(opt, defOpt)
      if phrase.find(word)==-1:
        return 0.0
      phraseParts = phrase.split(opt['sep'])
      rel=0.0 #relationship value of word to pharse(as a float number between 0 and 1
      for part0 in phraseParts:
        for ch in sch:
          part=part0.replace(ch,' ')
        pRel = 0 # part relationship
        if opt['matchWord']:
           pNum = 0
           partWords = takeStrWords(part)
           pLen = len(partWords)
           if pLen==0:
             continue
           for pw in partWords:
             if pw==word:
               pNum += 1
           pRel=float(pNum)/pLen  # part relationship
        else:
           pLen = len(part.replace(' ',''))
           if pLen==0:
             continue
           pNum = len(findAll(part, st)*len(st))
           pRel=float(pNum)/pLen  # part relationship
        if pRel > rel:
          rel = pRel
        #del pRel
      return rel

def charDigToInt(ch):
  if type(ch)!=str:
    raise RuntimeError, 'bad argument given to charDigToInt: "%s" must be a one length string.'%ch
  elif len(ch)!=1:
    raise RuntimeError, 'bad argument given to charDigToInt: "%s" must be a one length string.'%ch
  elif '0'<=ch<='9':
    return ord(ch)-48 #int(ch)
  elif 'a'<=ch<='z':
    return ord(ch)-87
  elif 'A'<=ch<='Z':
    return ord(ch)-55
  else:
    raise RuntimeError, 'bad argument given to charDigToInt: "%s" must be a one length string.'%ch
    

def intToBinStr(n, stLen=0):
  if type(stLen)!=int:
    raise TypeError, 'bad type second argument given to intToBinStr: "%s"'%type(stLen)
  h=hex(n)
  if h[-1]=='L':
    h = h[2:-1]
  else:
    h = h[2:]
  if len(h)%2==1:
    h='0'+h
  bs = ''
  for i in range(0, int(len(h)), 2):
    bs += chr( 16*charDigToInt(h[i]) + charDigToInt(h[i+1]) )
  bsl = len(bs)
  if bsl<stLen:
    bs = '\x00'*(stLen-bsl) + bs
  return bs

def binStrToInt(bs):
  l = len(bs)
  return sum([ ord(bs[i]) * (256**(l-1-i)) for i in range(l) ])
  """
  n = 0
  i = 0
  while i<l:
    #if bs[i]==' ':
    #  i += 1
    #elif bs[i]=='\\':
    #else:
      n += ord(bs[i]) * (256**(l-1-i))
      i += 1
  return n
  """


def intToIndexStr(n):
  # convert n to base 16 . create list "digits"
  digits = chBaseIntToList(n, 64)
  st=''
  for k in digits:
    if 0 <= k < 26:
      st += chr(65+k)
    elif 26 < k < 64:
      st += chr(71+k)
    else:
      raise RuntimeError, 'bad 16 based digit: %d'%k
  return st

def chBaseIntToStr(number, base):
    """intToStr( number, base ) -- reverse function to int(str,base) and long(str,base)"""
    if not 2 <= base <= 36:
      raise ValueError, "base must be in 2..36"
    abc = string.digits + string.letters
    result = ''
    if number < 0:
      number = -number
      sign = '-'
    else:
      sign = ''
    while True:
      number, rdigit = divmod( number, base )
      result = abc[rdigit] + result
      if number == 0:
         return sign + result

def chBaseIntToList(number, base):
    result = []
    if number < 0:
      raise ValueError, 'number must be posotive integer'
    while True:
      number, rdigit = divmod( number, base )
      result = [rdigit] + result
      if number == 0:
         return result


def recodeToWinArabic(s):
  u = s.decode('utf8')
  replaceList=[(u'ی',u'ي'),(u'ک',u'ك'),(u'ٔ',u'ء'),('\xef\xbf\xbd','')]+[(unichr(i),unichr(i+144)) for i in range(1632,1642)]
  for item in replaceList:
    u = u.replace(item[0], item[1])
  try:
    return u.encode('windows-1256')
  except:
    printAsError('can not encode string "%s" into windows-1256(arabic windows)'%s)
    return ''

def fileCopy(src, tar):
  os.popen("cp '%s' '%s'"%(src, tar))

  


def urlToPath(url):
  if len(url)<7:
    return url
  if url[:7]!='file://':
    return url
  path=url[7:]
  if path[-2:]=='\r\n':
    path=path[:-2]
  elif path[-1]=='\r':
    path=path[:-1]
  ## here convert html unicode symbols to utf8 string
  if not '%' in path:
    return path
  path2=''
  n=len(path)
  i=0
  while i<n:
    if path[i]=='%' and i<n-2:
      exec('path2 += chr(0x%s)'%path[i+1:i+3])
      i += 3
    else:
      path2 += path[i]
      i += 1
  return path2


faSubs={
u'،':[u'،',1 ,'fa_colon'	,0x22],
u'ـ':[u'ـ',1 ,'fa_underline'	,0xde],##	0xde
u'‌' :[u'‌', 1, 'virtual space'	,None],##	None
u'ی':[u'ﯼ',4 ,'fa_yeh'		,0xd6],
u'ا':[u'ﺍ',2 ,'alif'		,0x66],
u'إ':[u'ﺇ',2 ,'alif_down_hamzeh',0xdc],##	0x66	0xdc
u'آ':[u'ﺁ',2 ,'alif_madi'	,0x68],
u'أ':[u'ﺃ',2 ,'alif_up_hamzeh'	,0xda],##	0x66	0xda
u'ب':[u'ﺏ',4 ,'beh'		,0x6a],
u'پ':[u'ﭖ',4 ,'peh'		,0x6e],
u'ة':[u'ﺓ',2 ,'teh_tanith'	,0xe4],##	None	0xe4
u'ت':[u'ﺕ',4 ,'teh'		,0x72],
u'ث':[u'ﺙ',4 ,'theh'		,0x76],
u'ج':[u'ﺝ',4 ,'jeh'		,0x7a],
u'چ':[u'ﭺ',4 ,'cheh'		,0x7e],
u'ح':[u'ﺡ',4 ,'heh_jim'		,0x82],
u'خ':[u'ﺥ',4 ,'kheh'		,0x86],
u'د':[u'ﺩ',2 ,'dal'		,0x8a],
u'ذ':[u'ﺫ',2 ,'zal'		,0x8c],
u'ر':[u'ﺭ',2 ,'reh'		,0x8e],
u'ز':[u'ﺯ',2 ,'zeh'		,0x90],
u'ژ':[u'ﮊ',2 ,'zheh'		,0x92],
u'س':[u'ﺱ',4 ,'sin'		,0x94],
u'ش':[u'ﺵ',4 ,'shin'		,0x98],
u'ص':[u'ﺹ',4 ,'sad'		,0x9c],
u'ض':[u'ﺽ',4 ,'zad'		,0xa0],
u'ط':[u'ﻁ',4 ,'ta'		,0xa4],
u'ظ':[u'ﻅ',4 ,'za'		,0xa8],
u'ع':[u'ﻉ',4 ,'ein'		,0xac],
u'غ':[u'ﻍ',4 ,'ghein'		,0xb0],
u'ف':[u'ﻑ',4 ,'feh'		,0xb4],
u'ق':[u'ﻕ',4 ,'ghaf'		,0xb8],
u'ك':[u'ﻙ',4 ,'ar_kaf'		,0xda],##	0xbc	0xda
u'ک':[u'ﮎ',4 ,'fa_kaf'		,0xbc],
u'گ':[u'ﮒ',4 ,'gaf'		,0xc0],
u'ل':[u'ﻝ',4 ,'lam'		,0xc4],
u'م':[u'ﻡ',4 ,'mim'		,0xc8],
u'ن':[u'ﻥ',4 ,'noon'		,0xcc],
u'ه':[u'ﻩ',4 ,'heh'		,0xd2],
u'و':[u'ﻭ',2 ,'vav'		,0xd0],
u'ؤ':[u'ﺅ',2 ,'vav_up_hamzeh'	,0xde],##	0xd0	0xde
u'ي':[u'ﻱ',4 ,'ar_yeh'		,0xe0],##	0xd6	0xe0
u'ئ':[u'ﺉ',4 ,'yeh_up_hamzeh'	,0x62],
u'»':[u'»',1,'fa_quote_1'	,0xd8],
u'«':[u'«',1,'fa_quote_2'	,0xd9] ##	None	0xd9
}

faHarakat=[u'ّ', u'َ', u'ِ', u'ُ', u'ً', u'ٍ', u'ٌ', u'ْ', u'ٔ', u'ٓ']

def pmdCompile(uni):
  for h in faHarakat:
    uni=uni.replace(h, u'')
  missing=[]
  pmd=''
  en = False
  enStack=u''
  n=len(uni)
  for i in range(n):
    ch=uni[i]
    if u'A' <= ch <= u'z' or u'0'<= ch <=u'9':
      enStack = ch + enStack
      continue
    if enStack!=u'':
      pmd+=pmdEnConv(enStack)
      enStack=u''
    if not ch in faSubs.keys():
      pmd+=pmdEnConv(ch)
      continue
    stickPrev=False
    stickNext=False
    if uni[i] in faSubs.keys():
      if i>0 and uni[i-1] in faSubs.keys() and faSubs[uni[i]][1]>1:
        if faSubs[uni[i-1]][1]>3:
          stickPrev=True
      if i<n-1:
        if uni[i+1] in faSubs.keys() and faSubs[uni[i]][1]>3:
          if faSubs[uni[i+1]][1]>1:
            stickNext=True
    first=faSubs[ch][3]
    if first==None:
      missing.append(ch)
      continue
    if stickPrev and stickNext:          #medial  -> substitution 3
      pmd += chr(first+3)
    elif stickPrev and not stickNext:    #final   -> substitution 1
      pmd += chr(first+1)
    elif not stickPrev and stickNext:    #initial -> substitution 2
      pmd += chr(first+2)
    elif not stickPrev and not stickNext:#isolated-> substitution 0
      pmd += chr(first)
    del stickPrev, stickNext, first
  """
  if missing!=[]:
    fp=file('%s%smissing.txt'%(homeDir, os.sep), 'a')
    for ch in missing:
      print ch.encode('utf-8')
      #fp.write('%s\n'%ch.encode('utf-8'))
  """
  return pmd

pmdChars=[None,' ','~','`','!','@','#','$','%','^','&','*','-','_','+','=','\\','|','/','?',')','(',']','[','}','{','>','<',',','.',':',';','\'','"','“','،','?']

def pmdEnConv(uni):#gets an unicode(utf-16) string
  pmd=''
  s=uni.encode('utf-8')
  for c in s:
    if   'a'<=c<='z':
      pmd += chr(ord(c)-35)
    elif 'A'<=c<='Z':
      pmd += chr(ord(c)-29)
    elif '0'<=c<='9':
      pmd += chr(ord(c)+40)
    else:
      found=False
      for i in range(len(pmdChars)):
        if pmdChars[i]==c:
          pmd += chr(i)
          found=True
          break
      if not found:
        print 'no PMD coding for "%s"'%c
  return pmd

def pmdDecomile(pmd):
  uni=u''
  for c in pmd:
    for sub in list(faSubs):
      if sub[1][3] <= ord(c) < sub[1][3]+sub[1][1]:
        uni+=sub[0]
        break

