#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys, string, gzip


bgl_language = [
	"English", 
	"French",
	"Italian",
	"Spanish",
	"Dutch",
	"Portuguese",
	"German",
	"Russian",
	"Japanese",
	"Traditional Chinese",
	"Simplified Chinese",
	"Greek",
	"Korean",
	"Turkish",
	"Hebrew",
	"Arabic",
	"Thai",
	"Other",
	"Other Simplified Chinese dialects",
	"Other Traditional Chinese dialects",
	"Other Eastern-European languages",
	"Other Western-European languages",
	"Other Russian languages",
	"Other Japanese languages",
	"Other Baltic languages",
	"Other Greek languages",
	"Other Korean dialects",
	"Other Turkish dialects",
	"Other Thai dialects",
	"Polish",
	"Hungarian",
	"Czech",
	"Lithuanian",
	"Latvian",
	"Catalan",
	"Croatian",
	"Serbian",
	"Slovak",
	"Albanian",
	"Urdu",
	"Slovenian",
	"Estonian",
	"Bulgarian",
	"Danish",
	"Finnish",
	"Icelandic",
	"Norwegian",
	"Romanian",
	"Swedish",
	"Ukrainian",
	"Belarusian",
	"Farsi",
	"Basque",
	"Macedonian",
	"Afrikaans",
	"Faeroese",
	"Latin",
	"Esperanto",
	"Tamazight",
	"Armenian"]


bgl_charsetname = [
	"Default" ,
	"Latin",
	"Eastern European",
	"Cyrillic",
	"Japanese",
	"Traditional Chinese",
	"Simplified Chinese",
	"Baltic",
	"Greek",
	"Korean",
	"Turkish",
	"Hebrew",
	"Arabic",
	"Thai" ]

bgl_charset = [
	"ISO-8859-1", #Default
	"ISO-8859-1", #Latin
	"ISO-8859-2", #Eastern European
	"ISO-8859-5", #Cyriilic
	"ISO-8859-14", #Japanese
	"ISO-8859-14", #Traditional Chinese
	"ISO-8859-15", #Simplified Chinese
	"CP1257", #Baltic
	"CP1253", #Greek
	"ISO-8859-15",  #Korean
	"ISO-8859-9", #Turkish
	"ISO-8859-9", #Hebrew
	"CP1256", #Arabic
	"CP874"]  #Thai

partOfSpeech = [
  "n.",
  "adj.",
  "v.",
  "adv.",
  "interj.",
  "pron.",
  "prep.",
  "conj.",
  "suff.",
  "pref.",
  "art." ]


class bgl_block:
  def __init__(self):
    self.length=0
    self.data=''
    self.Type=''
  def __str__(self):
    return 'bgl_block Type=%s, length=%s, len(data)=%s'%(self.Type, self.length, len(self.data))

class bgl_entry:
  def __init__(self):
    self.headword=''
    self.definition=''
    self.alternate=''
    self.alternates=[]

DEFAULT_CHARSET=0
SOURCE_CHARSET =1
TARGET_CHARSET =2


class Babylon:
  def __init__(self, filename):
    self.m_filename = filename
    self.m_title=''
    self.m_author=''
    self.m_email=''
    self.m_description=''
    self.m_copyright=''
    self.m_sourceLang=''
    self.m_targetLang=''
    self.m_charset=''
  def open(self):
    f = file(str(self.m_filename))
    if f==None:
      return False
    buf = f.read(6)
    if len(buf)<6 or not buf[:4] in ('\x12\x34\x00\x01','\x12\x34\x00\x02'):
      return False
    i = ord(buf[4]) << 8 | ord(buf[5])
    print 'position of gz header: i=%s'%i
    if i<6:
      return False
    f.seek(i)
    f.flush()
    f2=file(self.m_filename+'-data.gz', 'w')
    f2.write(f.read())
    f.close()
    f2.close()
    self.file = gzip.open(self.m_filename+'-data.gz')
    return True
  def close(self):
    self.file.close()
  def readBlock(self, block):
    block.length = self.bgl_readnum( 1 )
    if block.length==False:
      return False
    block.Type = block.length & 0xf
    if block.Type == 4:
      return False
    block.length >>= 4
    if block.length < 4:
      block.length = self.bgl_readnum(block.length+1)
    else:
      block.length -= 4
    if block.length:
      try:
        block.data = self.file.read(block.length)
      except:
        print 'end of file.'
        sys.exit(0)
    return True
  def bgl_readnum(self, bytes):
    val=0
    if bytes<1 or bytes>4:
      return 0
    buf = self.file.read(bytes)
    if len(buf)==0:
      return False
    if len(buf)!=bytes:
      print 'bytes=%s , len(buf)=%s'%(bytes, len(buf))
      return False
    for i in range(bytes):
      val= (val << 8) | ord(buf[i])
    return val
  def read(self):
    self.m_numEntries = 0
    block=bgl_block()
    for ind in range(10000):
      self.readBlock(block)
      #print block
      headword=''
      definition=''
      if block.Type==0:
        if ord(block.data[0])==8:
          try:
            Type = ord(block.data[2])
          except:
            print block.data
            continue
          if Type > 64:
            Type -= 65
          self.m_defaultCharset = bgl_charset[Type]
      elif block.Type in (1,10,13):
        self.m_numEntries += 1
      elif block.Type==3:
        pos=2
        x = ord(block.data[1])
        if x==1:
          for a in range(block.length-2):
            headword += block.data[pos]
            pos += 1
          self.m_title = headword
        elif x==2:
          for a in range(block.length-2):
            headword += block.data[pos]
            pos += 1
          self.m_author = headword
        elif x==3:
          for a in range(block.length-2):
            headword += block.data[pos]
            pos += 1
          self.m_email = headword
        elif x==4:
          for a in range(block.length-2):
            headword += block.data[pos]
            pos += 1
          self.m_copyright = headword
        elif x==7:
          headword = bgl_language[ord(block.data[5])];
          self.m_sourceLang = headword
        elif x==8:
          headword = bgl_language[ord(block.data[5])];
          self.m_targetLang = headword
        elif x==9:
          for a in range(block.length-2):
            headword += block.data[pos]
            pos += 1
          self.m_description = headword
        elif x==26:
          headword = bgl_charset[ord(block.data[5])];
          self.m_sourceCharset = headword
        elif x==27:
          headword = bgl_charset[ord(block.data[5])];
          self.m_targetCharset = headword
        else:
          print 'ord(block.data[1])=%s'%x
      else:
        print 'block.Type=%s'%block.Type
        pass
    self.file.seek(0)
    self.m_title = self.convertToUtf8(self.m_title, DEFAULT_CHARSET)
    self.m_author = self.convertToUtf8(self.m_author, DEFAULT_CHARSET)
    self.m_email = self.convertToUtf8(self.m_email, DEFAULT_CHARSET)
    self.m_copyright = self.convertToUtf8(self.m_copyright, DEFAULT_CHARSET)
    self.m_description = self.convertToUtf8(self.m_description, DEFAULT_CHARSET)
    return True
  def readEntry(self):
    entry = bgl_entry()
    if self.file==None:
      entry.headword = ''
      entry.definition = ''
      return entry
    block = bgl_block()
    for ind in range(10000):
      self.readBlock(block)
      #print block
      if block.Type in (1,10,13):
        headword=''
        definition=''
        alternate=''
        alternates=[]
        temp=''
        pos = 0
        # Headword:
        Len = ord(block.data[pos]) ; pos+=1
        headword += block.data[pos:pos+Len]
        pos += Len
        headword = self.convertToUtf8(headword, SOURCE_CHARSET)
        #Definition:
        Len = ord(block.data[pos]) << 8; pos += 1
        Len |= ord(block.data[pos]); pos += 1
        for a in range(Len):
          try:
            block.data[pos]
          except:
            break
          if block.data[pos]=='\x0a':
            definition += '\n'
            pos += 1
          elif block.data[pos]<'\x20':
            if a<Len-3 and block.data[pos]=='\x14' and block.data[pos+1]=='\x02':
              definition = partOfSpeech[ord(block.data[pos+2]) - 0x30] + " " + definition
            pos += Len - a
          else:
            definition += block.data[pos] ; pos += 1
        definition = self.convertToUtf8(definition, TARGET_CHARSET)
        while pos < block.length:
          Len = ord(block.data[pos]); pos += 1
          alternate += block.data[pos:pos+Len]
          pos += Len
          alternate = self.convertToUtf8(alternate, SOURCE_CHARSET)
          alternates.append(alternate)
          alternate=''
        entry.headword = headword
        entry.definition = definition
        entry.alternates = alternates
        return entry
    entry.headword = ''
    return entry
  def convertToUtf8(self, s, Type):
    return s
    if s=='':
      return ''
    if Type>2:
      return s
    if Type==DEFAULT_CHARSET:
      if len(self.m_defaultCharset)>0:
        charset = self.m_defaultCharset
      else:
        charset = self.m_sourceCharset
    elif Type==SOURCE_CHARSET:
      if len(self.m_sourceCharset)>0:
        charset = self.m_sourceCharset
      else:
        charset = self.m_defaultCharset
    elif Type==TARGET_CHARSET:
      if len(self.m_targetCharset)>0:
        charset = self.m_targetCharset
      else:
        charset = self.m_defaultCharset



