#!/usr/bin/python
# -*- coding: utf-8 -*-
##   glossary.py 
##
##   Author: Saeed Rasooli <saeed.gnu@gmail.com>  (ilius)
##   Thanks to 'Mehdi Bayazee' and 'carp3(Pedram Azimaie)' for program bgl2xdb.py
##
##   This program is a free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 3, or (at your option)
##   any later version.
##
##   You can get a copy of GNU General Public License along this program
##   But you can always get it from http://www.gnu.org/licenses/gpl.txt
##
##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.

import string, os, sys, time

from text_utils import *
from interface import *

import warnings
warnings.resetwarnings()

#from bgl2x import BGL					## used for readBgl_int()
#import _babylon, _mdicbuilder, _stardictbuilder, _stardict
#from xml.sax.saxutils import XMLGenerator		## used for writeXfardic_2
#from xml.sax.xmlreader import AttributesNSImpl		## used for writeXfardic_2
#from xml.etree.ElementTree import XML, tostring	## used for readXfardic_2

cmpWord = lambda t1, t2: cmp(t1[0], t2[0])

class Glossary:
  tabfileExt=('.txt', '.tab', '.dic')
  infoKeysXfardic =['dbname','author','inputlang','version']+['outputlang','copyright','description']
  infoKeysStardict=['bookname', 'author','email','website','description','copyright','sourceLang','targetLang','charset']
  infoKeysSynonyms=[
    ['name'      , 'title'     , 'dbname'   , 'bookname'],
    ['sourceLang', 'inputlang' , 'origlang'],
    ['targetLang', 'outputlang', 'destlang'],
    ['copyright' , 'license'   ]]
  infoKeys = [g[0] for g in infoKeysSynonyms]
  readFormats=(
    'Tabfile'	,
    'Stardict'	,
    'Xfardic'	,
    'Bgl'	,
    'Mdic'	,
    'Xdxf'	,
    'Omnidic'  )
  writeFormats=(
    'Tabfile'	,
    'Stardict'	,
    'Xfardic'	,
    'Mdic'	,
    'Omnidic'	,
    'Pmd'	,
    'Dicformids',
    'Index'	,
    'Sql'	,
    'Freedict'	,
    'Babylon'	,
    'Sdict'	,
    'GettextPo'	,
    'Treedict'	)
  formatsDesc={
    'Tabfile'	:'Tabfile (txt,dic)',
    'Stardict'	:'StarDict (ifo)',
    'Xfardic'	:'xFarDic (xdb)',
    'Bgl'	:'Babylon (bgl)',
    'Bdc'	:'Babylon (bdc)',
    'Xdxf'	:'XDXF',
    'Mdic'	:'MDic (m2) (SQLite)',
    'Index'	:'Index',
    'Babylon'	:'Babylon Source (gls)',
    'Sdict'	:'Sdictionary Source(sdct)',
    'Sql'	:'SQL',
    'Freedict'	:'FreeDict (tei)',
    'GettextPo'	:'Gettext (po)',
    'GettextMo'	:'Gettext (mo)',
    'Omnidic'	:'Omnidic',
    'Pmd'	:'PMD',
    'Treedict'	:'TreeDict',
    'Dicformids':'DictionaryForMIDs'}
  formatsExt={
    'Tabfile'	:('.txt','.tab','.dic'),
    'Stardict'	:('.ifo',),
    'Xfardic'	:('.xdb',),
    'Bgl'	:('.bgl',),
    'Bdc'	:('.bdc',),
    'Xdxf'	:('.xdxf',),
    'Mdic'	:('.m2',),
    'Index'	:('.index',),
    'Babylon'	:('.gls','.babylon'),
    'Sdict'	:('.sdct',),
    'Sql'	:('.sql',),
    'Freedict'	:('.tei',),
    'GettextPo'	:('.po',),
    'GettextMo'	:('.mo',),
    'Omnidic'	:('.omni','.omnidic'),
    'Pmd'	:('.pmd',),
    'Treedict'	:('.tree','.treedict'),
    'Dicformids':('.mids',)}
  readExt   =[formatsExt[f] for f in readFormats]
  writeExt  =[formatsExt[f] for f in writeFormats]
  readDesc  =[formatsDesc[f] for f in readFormats]
  writeDesc =[formatsDesc[f] for f in writeFormats]
  descFormat=dict([ (formatsDesc[f],f) for f in formatsDesc.keys() ])
  descExt   =dict([ (formatsDesc[f],formatsExt[f][0]) for f in formatsDesc.keys() ])
  #extFormats=dict([ (formatsExt[f],f) for f in formatsExt.keys() ])
  def __init__(self, info=[], data=[]):
    self.info=[]
    self.setInfos(info, True)
    self.data = data

  __str__ = lambda self: 'glossary.Glossary'

  def __getattribute__(self, name):
    exec('atr = self.%s'%name)
    return atr

  def copy(self):
    return Glossary(self.info[:], self.data[:])

  def infoKeys(self):
    return [ t[0] for t in self.info ]

  def getInfo(self, key):
    lkey = str(key).lower()
    for group in Glossary.infoKeysSynonyms:
      if not type(group) in (list, tuple):
        raise TypeError, 'group=%s'%group
      if (key in group) or (lkey in group):
        #print 'key %s is in group %s'%(key, group)
        for skey in group:
          for t in self.info:
            if t[0]==skey:
              return t[1]
    for t in self.info:
      if t[0]==key or t[0].lower()==lkey:
        return t[1]
    return ''

  def setInfo(self, key, value):
    lkey = str(key).lower()
    for group in Glossary.infoKeysSynonyms:
      if not type(group) in (list, tuple):
        raise TypeError, 'group=%s'%group
      if (key in group) or (lkey in group):
        skey=group[0]
        for i in range(len(self.info)):
          if self.info[i][0]==skey:
            self.info[i]=(self.info[i][0],value)
            return
        for i in range(len(self.info)):
          if self.info[i][0] in group:
            self.info[i]=(self.info[i][0],value)
            return
    for i in range(len(self.info)):
      if self.info[i][0]==key or self.info[i][0].lower()==lkey:
          self.info[i]=(self.info[i][0],value)
          return
    self.info.append((key,value))
  def setInfos(self, info, setAll=False):
    for t in info:
      self.setInfo(t[0], t[1])
    if setAll:
      for key in self.infoKeys():
        if self.getInfo(key)=='':
          self.setInfo(key, '')

  def removeTags(self, removeBetween=False, tags=['i','I','b','B','a','A','tr','font','FONT','TR','td','TD']):
    n=len(self.data)
    for i in range(n):
      self.data[i] = (self.data[i][0], removeTextTags(self.data[i][1], removeBetween, tags))

  def read(self, filename, format='', ext=''):
    delFile=False
    if ext=='':
        ext=os.path.splitext(filename)[-1]
    ext = ext.lower()
    if ext in ('.gz','.bz2','.zip'):
        if ext=='.bz2':
          try:
            os.popen("bzip2 -dk '%s'"%filename)
          except:
            printAsError('Could not decompress file "%s"'%filename)
            return False
          else:
            filename = filename[:-4]
            ext=os.path.splitext(filename)[-1]
            delFile=True
        elif ext=='.gz':
          try:
            os.popen("gzip -dc '%s' > '%s'"%(filename,filename[:-3]))
          except:
            printAsError('Could not decompress file "%s"'%filename)
            return False
          else:
            filename = filename[:-3]
            ext=os.path.splitext(filename)[-1]
            delFile=True
        elif ext=='.zip':
          try:
            os.popen("unzip '%s' -d '%s'"%(filename,os.path.dirname(filename)))
          except:
            printAsError('Could not decompress file "%s"'%filename)
            return False
          else:
            filename = filename[:-4]
            ext=os.path.splitext(filename)[-1]
            delFile=True
    if format=='':
      for key in Glossary.formatsExt.keys():
        if ext in Glossary.formatsExt[key]:
          format = key
      if format=='':
          #if delFile:
          #  os.remove(filename)
          printAsError('Unknown extention "%s" for read support!'%ext)
          return False
    exec('self.read%s(filename=filename)'%format)
    if delFile:
      os.remove(filename)
    return True

  def write(self, filename='', format='', ext='',zipExt=''):
    if os.path.isdir(filename):
      if ext=='':
        if format=='':
          ext='.txt'
        else:
          ext=Glossary.formatsExt[format]
      filename = os.path.join(filename,os.path.split(self.getInfo('filename'))[-1]+ext)
    #elif os.path.exists(filename):
    #  printAsError('file "%s" already exists!'%filename)
    #  return False
    else:
      if ext=='':
        ext=os.path.splitext(filename)[-1]
      if ext in ('.gz','.bz2','.zip'):
          zipExt=ext
          filename = filename[:-len(zipExt)]
          ext=os.path.splitext(filename)[-1]
      if format=='':
        for key in Glossary.formatsExt.keys():
          if ext in Glossary.formatsExt[key]:
            format = key
        if format=='':
          printAsError('Unknown extention "%s" for write support!'%ext); return False

    exec('self.write%s(filename=filename)'%format)
    if zipExt in ('.gz','.bz2','.zip'):
      try:
        os.remove('%s%s'%(filename,zipExt))
      except OSError:
        pass
      if zipExt=='.gz':
        os.popen("gzip '%s'"%filename)
      elif zipExt=='.bz2':
        os.popen("bzip2 '%s'"%filename)
      elif zipExt=='.zip':
        (dirn,name)=os.path.split(filename)
        cwd=os.getcwd()
        os.chdir(dirn)
        try:
          os.popen("zip '%s.zip' '%s' -m"%(filename,name))
        except:
          printAsError('Could not compress with ".zip"')
        os.chdir(cwd)


  def readBgl_int1(self, filename):
    from bgl2x import BGL
    self.data = []
    BglConverter = BGL(filename)
    counter = 0
    if not BglConverter.open():
      raise IOError, 'Error on Reading bgl file !'
    BglConverter.ReadBglHeader()
    name = os.path.splitext(filename)[0]
    name = os.path.split(name)[1]
    while True:
        data = BglConverter.readWord()
        # Error or EOF
        if data == False :
          break
        elif data == None or data[0] == '' or data[1] == '': 
          continue
        #print  "[",data[0] , ']\n  [', data[1], "]\n==================\n"
        self.data += [data]
        counter += 1
    BglConverter.close()
    if filename[-4:].lower()=='.bgl':
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)


  def readBgl(self, filename):
    try:
      import _babylon
      from _babylon import Babylon_readEntry, bgl_entry_headword_get, bgl_entry_definition_get
    except ImportError:
      printAsError('Binary module "_babylon" can not be imported! Using internal BGL reader (this has not a good support)')
      return self.readBgl_int(filename)
    self.data = []
    db = _babylon.new_Babylon(filename)
    _babylon.Babylon_swigregister(db)
    if not _babylon.Babylon_open(db):
      raise 'can not open BGL file "%s"'%filename
    #print _babylon.Babylon_read(db)
    if not _babylon.Babylon_read(db):
      raise 'can not read BGL file "%s"'%filename
    n=_babylon.Babylon_numEntries(db)
    _babylon.Babylon_swigregister(n)
    if type(n)==int:
      print 'Babylon_numEntries = %s'%n
    ## could not determine numEntries as int when defined as "uint" in C++ codes. I changed its type to "int"
    try:
      self.ui
    except AttributeError:
      self.ui = None
    else:
      pass
    ui = self.ui
    if type(n)!=int:
      ui = None
    if ui==None:
      while True:
        entry = Babylon_readEntry(db)
        _babylon.Babylon_swigregister(entry)
        w = bgl_entry_headword_get(entry)
        m = bgl_entry_definition_get(entry).replace('<BR>','\n').replace('<br>','\n')
        if w=='' and m=='':
          break
        else:
          self.data.append((w,m))
    else:
      k = 2000
      for i in range(n):
        entry = Babylon_readEntry(db)
        _babylon.Babylon_swigregister(entry)
        w = bgl_entry_headword_get(entry)
        m = bgl_entry_definition_get(entry).replace('<BR>','\n').replace('<br>','\n')
        self.data.append((w,m))
        if i%k==0:
          rat = float(i)/n
          ui.progress(rat)
          while gtk.events_pending():
            gtk.main_iteration_do(False)
      ui.progress(1.0, 'Loading Completed')
      while gtk.events_pending():
        gtk.main_iteration_do(False)
    self.setInfo('title'       ,_babylon.Babylon_title(db)      )
    self.setInfo('author'      ,_babylon.Babylon_author(db)     )
    self.setInfo('email'       ,_babylon.Babylon_email(db)      )
    self.setInfo('description' ,_babylon.Babylon_description(db))
    self.setInfo('copyright'   ,_babylon.Babylon_copyright(db)  )
    self.setInfo('sourceLang'  ,_babylon.Babylon_sourceLang(db) )
    self.setInfo('targetLang'  ,_babylon.Babylon_targetLang(db) )
    self.setInfo('charset'     ,_babylon.Babylon_charset(db)    )
    if filename[-4:].lower()=='.bgl':
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)


  def readBgl_int(self, filename):
    import bgl
    self.data = []
    db = bgl.Babylon(filename)
    if not db.open():
      raise 'can not open BGL file "%s"'%filename
    if not db.read():
      raise 'can not read BGL file "%s"'%filename
    n=db.m_numEntries
    if type(n)==int:
      print 'Babylon_numEntries = %s'%n
    try:
      self.ui
    except AttributeError:
      self.ui = None
    ui = self.ui
    if type(n)!=int:
      ui = None
    if ui==None:
      while True:
        entry = db.readEntry()
        new = (entry.headword, entry.definition)
        if new==('', ''):
          break
        else:
          self.data.append(new)
    else:
      k = 2000
      for i in range(n):
        entry = db.readEntry()
        self.data.append((entry.headword, entry.definition))
        if i%k==0:
          rat = float(i)/n
          ui.progress(rat)
          while gtk.events_pending():
            gtk.main_iteration_do(False)
      ui.progress(1.0, 'Loading Completed')
      while gtk.events_pending():
        gtk.main_iteration_do(False)
    """self.setInfo('title'       ,db.title()      )
    self.setInfo('author'      ,db.author()     )
    self.setInfo('email'       ,db.email()      )
    self.setInfo('description' ,db.description())
    self.setInfo('copyright'   ,db.copyright()  )
    self.setInfo('sourceLang'  ,db.sourceLang() )
    self.setInfo('targetLang'  ,db.targetLang() )
    self.setInfo('charset'     ,db.charset()    )"""
    if filename[-4:].lower()=='.bgl':
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)

  def writeBgl(self, filename=''):
    pass

  def readXfardic(self, filename):
    fp=file(filename,'r')
    self.data=[]
    xdbText=fp.read()
    i=0
    for item in Glossary.infoKeysXfardic():##########################   must be change method
      inf0 = xdbText.find('<'+item+'>', i)
      if inf0==-1:
        continue
      inf0 += (len(item)+2)
      inf1 = xdbText.find('</'+item+'>', inf0)
      inf = xdbText[inf0:inf1]
      self.setInfo(item, inf)
      i = inf1
    while True:######################################################################
      i=xdbText.find('<word>', i)
      if i==-1:
        break
      in0 = xdbText.find('<in>', i) + 4
      in1 = xdbText.find('</in>', in0)
      out0= xdbText.find('<out>', in1) + 5
      out1= xdbText.find('</out>', out0)
      word = xdbText[in0:in1]
      mean = xdbText[out0:out1]
      self.data.append((word,mean))
      #i = out1
      i = xdbText.find('</word>', out1) + 7    
    if filename[-4:].lower() in ('.xdb', '.xml'):
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('dbname') == '':
      dbname = os.path.split(filename)[-1]
      self.setInfo('dbname', dbname)


  def readXfardic_2(self, filename):
    from  xml.etree.ElementTree import XML, tostring
    fp=file(filename,'r')
    self.data=[]
    self.info={}
    xdb = XML(fp.read())
    del fp
    for elem in xdb[0]:
      et = tostring(elem)
      i0 = et.find('<')
      i1 = et.find('>',i0+1)
      i2 = et.find('<',i1+1)
      self.info[et[i0:i1]] = et[i1+1:i2]
    for elem in xdb[1:]:
      try:
        w,m = tostring(elem[0]), tostring(elem[1])
      except:
        print tostring(elem)
        continue
      self.data += [[ w[4:-5] , m[5:-6] ]]
    if filename[-4:].lower() in ('.xdb', '.xml'):
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('dbname') == '':
      dbname = os.path.split(filename)[-1]
      self.setInfo('dbname', dbname)


  def writeXfardic(self,filename=''):
    if filename=='':
      filename = self.getInfo('filename')+'.xdb'
    fp=file(filename,'w')
    fp.write('<?xml version="1.0" encoding="utf-8" ?>\n<words>\n<xfardic>')
    for item in Glossary.infoKeysXfardic:
      fp.write('<'+item+'>'+str(self.getInfo(item))+'</'+item+'>')
    fp.write('</xfardic>\n')
    for item in self.data:
      fp.write("<word><in>"+item[0]+"</in><out>"+ item[1]+"</out></word>\n")
    fp.write("</words>\n")
    fp.close()


  def writeXfardic_2(self, filename=''):
    from xml.sax.saxutils import XMLGenerator
    from xml.sax.xmlreader import AttributesNSImpl
    if filename=='':
      filename = self.getInfo('filename')+'.xdb'
    xdbFp=file(filename,'w')
    fp = XMLGenerator(xdbFp, 'utf-8')
    attrs = AttributesNSImpl({}, {})
    fp.startElement(u'xfardic', attrs)
    for t in self.info:
      fp.startElement(unicode(t[0]), attrs)
      fp.characters(unicode(t[1]))
      fp.endElement(unicode(t[0]))
    fp.endElement(u'xfardic')
    fp.startElement(u'words', attrs)
    for item in self.data:
      try:
        tmpXmlFile.characters(item[1])
      except:
        printAsError('While writing xdb file, an error on word "%s"'%item[0])
        continue
      fp.startElement(u'word', attrs)
      fp.startElement(u'in', attrs)
      fp.characters(unicode(item[0]))
      fp.endElement(u'in')
      fp.startElement(u'out', attrs)
      fp.characters(unicode(item[1]))
      fp.endElement(u'out')
    fp.endElement(u'words')
    fp.endDocument()
    xdbFp.close()



  def readTabfile(self, filename):
    fp=file(filename)
    self.data=[]
    lines=fp.readlines()
    for k in range(len(lines)):
      line = lines[k]
      #spl=line.rsplit("\t")
      #word=spl[0]
      #mean=spl[1][:-1] # to remove newline character
      fti=line.find('\t') # first tab's index
      if fti==-1:
        printAsError('Warning: line %s beganing "%s" has no tab!' %(k, line[:10]))
        continue
      word = line[:fti]
      mean = line[fti+1:-1]#.replace('\\n','\n')#.replace('<BR>','\n').replace('\\t','\t')
      for i in range(128):
        c = chr(i)
        if not c in mean:
          mean = mean.replace('\\\\n',c).replace('\\n','\n').replace(c,'\\n')
          mean = mean.replace('\\\\t',c).replace('\\t','\t').replace(c,'\\t')
          break
      if len(word)>0:
        if word[0]=='#':
          while word[0]=='#':
            word=word[1:]
            if len(word)==0:
              break
          self.setInfo(word, mean)
          continue
      self.data.append((word,mean))
    if filename[-4:].lower() in Glossary.tabfileExt:
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name')=='':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)


  def writeTxt(self, sep, filename='', writeInfo=False, rplList=[], ext='.txt', head=''):
    if filename=='':
      filename=self.getInfo('filename')+ext
    txt=head
    if writeInfo:
      for t in self.info:
        inf = self.getInfo(t[1])
        if inf!='':
          try:
            txt+=('##' + t[0] + sep[0] + inf + sep[1])
          except:
            printAsError('Error on writing info line for "%s"'%t[0])
    for item in self.data:
      (word,mean)=item
      if word[0]=='#':
        continue
      for rpl in rplList:
        mean=mean.replace(rpl[0],rpl[1])
      try:
        line = word+sep[0]+mean+sep[1]
      except:
        printAsError('The line for word "%s" is not unicode'%word)
        continue
      try:
        txt+=line
      except:
        printAsError('Error on writing line for word "%s"'%word)
    if filename==None:
      return txt
    fp=file(filename,'w')
    fp.write(txt)
    fp.close()
    return True

  def writeTabfile(self, filename='', writeInfo=True):
    return self.writeTxt(['\t','\n'], filename=filename, writeInfo=writeInfo, rplList=[['\n','\\n']], ext='.txt')



  def writeTabfile_2(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.txt'
    fp = file(filename,'w')
    xmlFile = XMLGenerator(fp, 'utf-8')
    attrs = AttributesNSImpl({}, {})
    for item in self.data:
      try:
        tmpXmlFile.characters(item[1])
      except:
        printAsError('While writing xdb. an error on word "%s"'%item[0])
        continue
      xmlFile.characters(unicode(item[0]))
      xmlFile.characters(u'\t')
      xmlFile.characters(unicode(item[1]))
      #xmlFile.characters(u'\n')
    xmlFile.endDocument()
    fp.close()



  def writeBabylon(self, filename='', writeInfo=True, newline='', encoding='utf8'):
    ## Source Glossary for "Babylon Builder".
    ## It has extention '.gls' or '.babylon'. But not '.bgl'.
    ## It is a (unicode) text file. Not binary like bgl files.
    winArabic=['windows-1256', 'windows-arabic', 'arabic-windows','arabic windows','windows arabic']
    g = self
    encoding = encoding.lower()
    if not encoding in ('utf8',''):
      g = self.copy()
      if encoding in winArabic:
        for i in range(len(g.data)):
          g.data[i] = (recodeToWinArabic(g.data[i][0]), recodeToWinArabic(g.data[i][1]))
        if newline=='':
          newline='\r\n'
      else:
        for i in range(len(g.data)):
          g.data[i]=(g.data[i][0].decode('utf8').encode(encoding), g.data[i][1].decode('utf8').encode(encoding))
    if newline=='':
      newline='\n'
    head=''
    if writeInfo:
      head += '### Glossary title:%s%s' %(g.getInfo('name'),newline)
      head += '### Author:%s%s'         %(g.getInfo('author'),newline)
      head += '### Description:%s%s'    %(g.getInfo('description'),newline)
      head += '### Source language:%s%s'%(g.getInfo('inputlang'),newline)
      if encoding in winArabic:
        head += '### Source alphabet:Arabic%s'%newline
      else:
        head += '### Source alphabet:%s%s'%(encoding,newline)
      head += '### Target language:%s%s'%(g.getInfo('outputlang'),newline)
      if encoding in winArabic:
        head += '### Target alphabet:Arabic%s'%newline
      else:
        head += '### Target alphabet:%s%s'%(encoding,newline)
      head += '### Browsing enabled?Yes%s'%newline
      head += '### Type of glossary:00000000%s'%newline
      head += '### Case sensitive words?0%s'%newline
      head += '\n### Glossary section:%s'%newline
    g.writeTxt([newline,newline*2], filename=filename, writeInfo=False, rplList=[['\n','<BR>']], ext='.gls', head=head)


  def writeSdict(self, filename='', writeInfo=True, newline='\n'):
    ## Source Glossary for "Sdictionary".
    ## It has extention '.sdct' 
    head=''
    if writeInfo:
      head += '<header>\n'
      head += 'title = %s\n' %self.getInfo('name')
      head += 'author = %s\n'         %self.getInfo('author')
      head += 'description = %s\n'    %self.getInfo('description')
      head += 'w_lang = %s\n'%self.getInfo('inputlang')
      head += 'a_lang = %s\n'%self.getInfo('outputlang')
      head += '</header>\n#\n#\n#\n'
    self.writeTxt(['___',newline], filename, False, [['\n','<BR>']], '.sdct', head)

  def writeDict(self, filename='', writeInfo=False): ## Uses in '/usr/share/dict/' for some dictionarys such as 'ding'.
    self.writeTxt([' :: ','\n'], filename, writeInfo, [['\n','\\n']], '.dict')



  def printTabfile(self):
    errors = 0
    for item in self.data:
      (word,mean)=item
      mean=mean.replace('\n','\\n')
      try:
        print word+'\t'+mean
      except:
        errors += 1
        continue
    print 'errors\t'+str(errors)

  def readXdxf(self, filename):
    ##<!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
    fp=file(filename,'r')
    xdxf=XML(fp.read())
    fp.close()
    full_name  =tostring(xdxf[0]).replace('<full_name>'  ,'').replace('</full_name>'  ,'')
    description=tostring(xdxf[1]).replace('<description>','').replace('</description>','')
    while full_name[-1]=='\n':
      full_name=full_name[:-1]
    while description[-1]=='\n':
      description=description[:-1]
    self.setInfo('name', full_name)
    self.setInfo('description',description)
    maxWordLen=0
    for item in xdxf[2:]:
      if len(item)==2:
        mean=tostring(item[1]).replace('<tr>','').replace('</tr>','\n')
        word=tostring(item[0]).replace('<k>','').replace('</k>','')
      elif len(item)==1:
        itemStr=tostring(item[0])
        ki = itemStr.find('</k>')
        mean = itemStr[ki+4:]
        word = itemStr[:ki].replace('<k>','')
      else:
        pass#print word, len(item)
      while word[-1]=='\n':
        word=word[:-1]
      if mean!='':
        while mean[-1]=='\n':
          mean=mean[:-1]
        while mean[0]=='\n':
          mean=mean[1:]
      wordLen=len(word) 
      if maxWordLen<len(word):
        maxWordLen=len(word)
      self.data.append((word,mean))
      del word, mean
    if filename[-4:].lower()=='.xml':
      filename = filename[:-4]
    elif filename[-5:].lower()=='.xdxf':
      filename = filename[:-5]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)
    
  ###################################################################
  def checkUnicode(self):
    n = len(self.data)
    i = 0
    while True:
      try:
        data[i]
      except:
        break
      try:
        unicode(self.data[i][0])
        unicode(self.data[i][1])
      except:
        printAsError('Line for word "%s" is not unicode! removing line.'%self.data[i][0])
        self.data.remove(i)
        continue
      i += 1

  takeWords = lambda self: [ item[0] for item in self.data ]


  def takeOutputWords(self, opt={}):
    words=takeStrWords(string.join([item[1] for item in self.data],' '),opt)
    words.sort()
    words=removeRepeats(words)
    return words

  def simpleSwap(self):
    return Glossary(self.info[:], [ (item[1],item[0]) for item in self.data ])

  def attach(self, other):# only simplicity attach two glossarys(or more that others be as a list).
  # no ordering. Use when you splited input words to two(or many) parts after ordering.
    try:
      other.data, other.info
    except:
      if type(other) in ( list, tupleType ):
        if len(other)==0:
          return self
        if len(other)==1:
          return self.attach(other[0])
        return self.attach(other[0]).attach(other[1:])
      else:
        return self
    newName = '"%s" attached to "%s"'%( self.getInfo('name') , other.getInfo('name') )
    ng=Glossary(  [('name',newName)]  , self.data + other.data)
    #################################         here attach and set info of two glossary
    return ng

  def merge(self, other):
    try:
      other.data, other.info
    except:
      if type(other) in ( list, tupleType ):
        if len(other)==0:
          return self
        if len(other)==1:
          return self.merge(other[0])
        return self.merge(other[0]).merge(other[1:])
      else:
        raise TypeError, 'bad argument given to merge! other="%s"'%other
    newName = '"%s" merged with "%s"'%( self.getInfo('name') , other.getInfo('name') )
    new = Glossary(  [('name',newName)]  )
    new.data = self.data + other.data
    new.data.sort()
    return new


  def deepMerge(self, other, sep="\n"):#merge two optional glossarys nicly. no repets in words of result glossary
    try:
      other.data, other.info
    except:
      if type(other) in ( list, tupleType ):
        if len(other)==0:
          return self
        if len(other)==1:
          return self.deepMerge(other[0])
        return self.deepMerge(other[0]).deepMerge(other[1:])
      else:
        raise TypeError, 'bad argument given to deepMerge! other="%s"'%other
    newName = '"%s" deep merged with "%s"'%( self.getInfo('name') , other.getInfo('name') )
    new = Glossary(  [('name',newName)]  )
    data = list(self.data + other.data)
    data.sort(cmpWord)
    n=len(data)
    i=0
    while i<len(data)-1:
      if data[i][0]==data[i+1][0]:
        if data[i][1]!=data[i+1][1]:
          data[i] = ( data[i][0], data[i][1]+sep+data[i+1][1] )
        data.pop(i+1)
      else:
        i += 1
    new.data=data
    return new


  def __add__(self, other):
    return self.merge(other) 



  def searchWordInDef(self, st, opt):
    #seachs word 'st' in meanings(definitions) of the glossary 'self'
    defOpt={'minRel':0.0, 'maxNum':100, 'sep':commaFa, 'matchWord':True, 'showRel':'Percent'}
    opt = addDefaultOptions(opt, defOpt)
    sep=opt['sep']
    matchWord=opt['matchWord']
    maxNum = opt['maxNum']
    outRel = []
    for item in self.data:
      (word,mean)=item[:]
      meanParts = mean.split(sep)
      if mean.find(st)==-1:
        continue
      rel=0 #relationship value of word (as a float number between 0 and 1
      for part in meanParts:
        for ch in sch:
          part=part.replace(ch,' ')
        pRel = 0 # part relationship
        if matchWord:
           pNum = 0
           partWords = takeStrWords(part)
           pLen = len(partWords)
           if pLen==0:
             continue
           for pw in partWords:
             if pw==st:
               pNum += 1
           pRel=float(pNum)/pLen  # part relationship
        else:
           pLen = len(part.replace(' ',''))
           if pLen==0:
             continue
           pNum = len(findAll(part, st))*len(st)
           pRel=float(pNum)/pLen  # part relationship
        if pRel > rel:
          rel = pRel
      if rel <= opt['minRel']:
        continue
      outRel.append((word[:],rel))
    sortby_inplace(outRel, 1, True)
    n=len(outRel)
    if n > maxNum > 0:
      outRel=outRel[:maxNum]
      n=maxNum
    num=0
    out=[]
    for j in range(n):
      numP=num
      (w,num)=outRel[j]
      #if type(w)!=str:
      #  raise TypeError, 'w=%s   type(w)=%s'%(w, type(w))
      onePer=int(1.0/num)
      if onePer==1.0:
        out.append(w)
      elif opt['showRel']=='Percent':
        out.append('%s(%%%d)' % ( w, 100*num ))
      elif opt['showRel']=='Percent At First':
        if num==numP:
          out.append(w)
        else:
          out.append('%s(%%%d)' % ( w, 100*num ))
      else:
        out.append(w)
      #for item in out:
      #  if type(item)!=str:
      #    raise TypeError, 'item=%s   type(item)=%s'%(item, type(item))
    return out


  def reverseDic_int(self, wordsArg=None, opt={}):
    defOpt={
    'matchWord':True,
    'showRel':'None',
    'background':False,
    'reportStep':300,
    'autoSaveStep':1000, ## set this to zero to disable auto saving.
    'savePath':''}
    opt = addDefaultOptions(opt, defOpt)
    self.stoped = False
    ui=self.ui
    try:
      c = self.continueFrom
    except AttributeError:
      c = 0
    if c==-1:
      print 'c=%s'%c
      return
    elif c==0:
      ui.progressbar.set_text('Starting....')
    if wordsArg==None:
      words = self.takeOutputWords()
    elif type(wordsArg)==file:
      words = [ w[:-1] for w in wordsArg.readlines() ]
    elif type(wordsArg)==list:
      words = wordsArg[:]
    elif type(wordsArg)==str:
      fp=file(wordsArg)
      words = [ w[:-1] for w in fp.readlines() ]
      fp.close()
    else:
      raise TypeError, 'Argumant wordsArg to function reverseDic is not valid!'
    autoSaveStep=opt['autoSaveStep']
    if opt['savePath']=='':
      opt['savePath']=self.getInfo('name')+'.txt'
    savePath = opt['savePath']
    if c > 0:
      saveFile = file(savePath, 'a')
    else:
      saveFile = file(savePath, 'w')
    join=string.join
    revG = Glossary(self.info[:])
    revG.setInfo('name', self.getInfo('name')+'_reversed')
    revG.setInfo('inputlang' , self.getInfo('outputlang'))
    revG.setInfo('outputlang', self.getInfo('inputlang'))
    wNum=len(words)
    #steps = opt['reportStep']
    #div = 0
    #mod = 0
    #total = int(wNum/steps)
    if c==0:
      print 'Number of input words:', wNum
      print 'Reversing glossary...'
    else:
      print 'continue reversing from index %d ...'%c
    t0=time.time()
    if ui==None:
      print 'passed ratio         time:  passed       remain          total       process'
    n = len(words)
    for i in range(c, n):
       word = words[i]
       rat = float(i+1)/n
       ui.progress(rat, '%d / %d words completed'%(i,n))
       if ui.reverseStop:
         saveFile.close()
         self.continueFrom = i
         self.stoped = True
         #thread.exit_thread()
         return
       while gtk.events_pending():
         gtk.main_iteration_do(False)
       """
       if mod == steps:
         mod = 0 ; div += 1
         t = time.time()
         dt = t-t0
         tRem = (total-div)*dt/div ## (n-i)*dt/n
         rat = float(i)/n
         if ui==None:
           print '%4d / %4d               %8s\t%8s\t%8s\t%s'%(div,total,timeHMS(dt),\
             timeHMS(tRem),timeHMS(dt+tRem),sys.argv[0])
         else:
           #############  ??????????????????????????????????????????????????????????????
           #ui.progressbar.set_text('%d/%d words completed (%%%2f) remaining %d seconds'%(i,n,rat*100,tRem))
           ui.progressbar.update(rat)
           while gtk.events_pending():
             gtk.main_iteration_do(False)
       else:
         mod += 1
       """
       if autoSaveStep>0 and i%autoSaveStep==0 and i>0:
         saveFile.close()
         saveFile = file(savePath, 'a')
       result = self.searchWordInDef(word, opt)
       if len(result)>0:
         try:
           new = ( word , join(result, ', ') + '.' )
         except:
           file('result','w').write(str(result));sys.exit(0)
         if autoSaveStep>0:
           saveFile.write('%s\t%s\n'%new)
         else:
          revG.data.append(new) 
       if autoSaveStep>0 and i==n-1:
         saveFile.close()
    if autoSaveStep==0:
      revG.writeTabfile(opt['savePath'])
    ui.r_finished()
    return True




  def reverseDic(self, wordsArg=None, opt={}):
    from _reverse_dic import search
    tabStr=self.writeTabfile(filename=None)
    defOpt={
    'matchWord':True,
    'showRel':'None',
    'background':False,
    'reportStep':300,
    'autoSaveStep':1000, ## set this to zero to disable auto saving.
    'savePath':'',
    'sep':commaFa}
    opt = addDefaultOptions(opt, defOpt)
    self.stoped = False
    ui=self.ui
    try:
      c = self.continueFrom
    except AttributeError:
      c = 0
    if c==-1:
      print 'c=%s'%c
      return
    elif c==0:
      ui.progressbar.set_text('Starting....')
    if wordsArg==None:
      words = self.takeOutputWords()
    elif type(wordsArg)==file:
      words = [ w[:-1] for w in wordsArg.readlines() ]
    elif type(wordsArg)==list:
      words = wordsArg[:]
    elif type(wordsArg)==str:
      fp=file(wordsArg)
      words = [ w[:-1] for w in fp.readlines() ]
      fp.close()
    else:
      raise TypeError, 'Argumant wordsArg to function reverseDic is not valid!'
    autoSaveStep=opt['autoSaveStep']
    if opt['savePath']=='':
      opt['savePath']=self.getInfo('name')+'.txt'
    savePath = opt['savePath']
    if c > 0:
      saveFile = file(savePath, 'a')
    else:
      saveFile = file(savePath, 'w')
    join=string.join
    revG = Glossary(self.info[:])
    revG.setInfo('name', self.getInfo('name')+'_reversed')
    revG.setInfo('inputlang' , self.getInfo('outputlang'))
    revG.setInfo('outputlang', self.getInfo('inputlang'))
    wNum=len(words)
    #steps = opt['reportStep']
    #div = 0
    #mod = 0
    #total = int(wNum/steps)
    if c==0:
      print 'Number of input words:', wNum
      print 'Reversing glossary...'
    else:
      print 'continue reversing from index %d ...'%c
    t0=time.time()
    if ui==None:
      print 'passed ratio         time:  passed       remain          total       process'
    n = len(words)
    for i in range(c, n):
       word = words[i]
       rat = float(i+1)/n
       ui.progress(rat, '%d / %d words completed'%(i,n))
       if ui.reverseStop:
         saveFile.close()
         self.continueFrom = i
         self.stoped = True
         #thread.exit_thread()
         return
       while gtk.events_pending():
         gtk.main_iteration_do(False)
       """
       if mod == steps:
         mod = 0 ; div += 1
         t = time.time()
         dt = t-t0
         tRem = (total-div)*dt/div ## (n-i)*dt/n
         rat = float(i)/n
         if ui==None:
           print '%4d / %4d               %8s\t%8s\t%8s\t%s'%(div,total,timeHMS(dt),\
             timeHMS(tRem),timeHMS(dt+tRem),sys.argv[0])
         else:
           #############  ??????????????????????????????????????????????????????????????
           #ui.progressbar.set_text('%d/%d words completed (%%%2f) remaining %d seconds'%(i,n,rat*100,tRem))
           ui.progressbar.update(rat)
           while gtk.events_pending():
             gtk.main_iteration_do(False)
       else:
         mod += 1
       """
       if autoSaveStep>0 and i%autoSaveStep==0 and i>0:
         saveFile.close()
         saveFile = file(savePath, 'a')
       result = search(tabStr, word, opt['minRel'], opt['maxNum'], opt['sep'], opt['matchWord'], opt['showRel'])
       if len(result)>0:
         new = ( word , result )
         if autoSaveStep>0:
           saveFile.write('%s\t%s\n'%new)
         else:
          revG.data.append(new) 
       if autoSaveStep>0 and i==n-1:
         saveFile.close()
    if autoSaveStep==0:
      revG.writeTabfile(opt['savePath'])
    ui.r_finished()
    return True



  def replaceInDefinitions(self, replaceList, matchWord=False):
    if not matchWord:
      for rpl in replaceList:
        for i in range(len(self.data)):
           if self.data[i][1].find(rpl[0])>-1:
             self.data[i] = (self.data[i][0], self.data[i][1].replace(rpl[0], rpl[1]))
    else:
      num=0
      for rpl in replaceList:
        for j in range(len(self.data)):
          # words indexes
          wdsIdx=findWords(self.data[j][1], {'word':rpl[0]})
          for [i0,i1] in wdsIdx:
            self.data[j][1] = self.data[j][1][:i0] + rpl[1] + self.data[j][1][i1:]
            num += 1
      return num

  def faEdit(self):
    for item in self.data:
      item[1] = item[1].replace('\xd9\x8a','\xdb\x8c') # ya
      item[1] = item[1].replace('\xd9\x83','\xda\xa9') # kaf
    # separate word "va" ('\xd9\x88') from next word 
    # if not part of the word.
    vav = '\xd9\x88'
    vavCorFp=file('vav_corrects.txt','r')
    vavCor = [ line[:-1] for line in vavCorFp.readlines() ]
    ## first way
    """
    for j in range(len(self.data)):
      for word in takeStrWords(self.data[j][1], {'minLen':6}):
        if word[:2]==vav and not word in vavCor:
          self.data[j][1] = self.data[j][1].replace(word, vav+' '+word[2:])"""
    ## second way (prefered)
    for j in range(len(self.data)):
      for [i0,i1] in findWords(self.data[j][1]):
        if self.data[j][1][i0:i0+2]==vav and not self.data[j][1][i0:i1] in vavCor:
          if i1-i0<6:
            print self.data[j][1][i0:i1]
            continue
          self.data[j][1] = self.data[j][1][:i0+2]+' '+self.data[j][1][i0+2:]


  def takePhonetic_oxford_gb(self):
   phg = Glossary(self.info[:]) # phonetic glossary
   phg.setInfo('name', self.getInfo('name')+'_phonetic')
   for item in self.data:
     word=item[0]
     mean=item[1]
     if mean[0]!="/":
       continue
     #### Now set the phonetic to the `ph` variable.
     ph=""
     sep=["/ adj", "/ v", "/ n", "/ adv", "/adj", "/v", "/n","/adv" ]+["/ n","/ adj", "/ the"]
     for s in sep:
       i = mean.find(s,2,85)
       if i==-1:
         continue
       else:
         ph=mean[:i+1]
         break
     ph=ph.replace(';','\t')
     ph=ph.replace(',','\t')
     #ph=ph.replace('/','')
     #ph=ph.replace('\\n ','\\n')
     #ph=ph.replace('\\n  ','\\n')
     ph=ph.replace('     ','\t')
     ph=ph.replace('    ','\t')
     ph=ph.replace('  ','\t')
     ph=ph.replace('	','\t')
     ph=ph.replace('//','/')
     ph=ph.replace('\t/\t','\t')
     ph=ph.replace('<i>US</i>\t','\tUS: ')
     ph=ph.replace('<i>US</i>','\tUS: ')
     ph=ph.replace('\t\t\t','\t')
     ph=ph.replace('\t\t','\t')
     if ph != "":
       phg.data.append((word, ph))
   return phg



  def writeOmnidic(self, filename='', dicIndex=16):
    if type(filename)!=str:
      raise TypeError, 'Invalid argument to function writeOmnidic: dicIndex=%s'%dicIndex
    if type(dicIndex)!=int:
      raise TypeError, 'Invalid argument to function writeOmnidic: filename=%s'%filename
    if filename=='':
      filename = self.getInfo('name')+'.omnidic'
      filename = filename.replace('/','_')
      filename = filename.replace('\\','_')
    if not os.path.isdir(filename):
      os.mkdir(filename)
    os.chdir(filename)
    wordCount=len(self.data)
    fileCountM1=int(wordCount/100) # file count mines one
    indexFp=file(str(dicIndex),'w')
    for i in range(fileCountM1):
      if i==0:
        fileName=str(dicIndex)+'99'
      else:
        fileName=str(dicIndex)+str(i)+'99'
      indexFp.write(self.data[100*i][0]+'#'+self.data[100*i+99][0]+'#'+fileName+'\n')
      fp=file(fileName,'w')
      for j in range(99):
        (w,m)=self.data[100*i+j]
        m=m.replace('\n','  ')  ######## ??????????????????????????????????????????????????????????????????
        fp.write('%s#%s\n'%(w,m))
      fp.write('%s#%s\n'%tuple(self.data[100*i+99][:2]))
      fp.close(); del fp, fileName
    endNum = wordCount % 100
    if endNum > 0:
      fileName=str(dicIndex)+str(fileCountM1)+str(endNum-1)
      fp=file(fileName,'w')
      indexFp.write(self.data[100*fileCountM1][0]+'#'+self.data[100*fileCountM1+endNum-1][0]+'#'+fileName+'\n')
      for j in range(endNum-1):
        fp.write(self.data[100*fileCountM1+j][0]+'#'+self.data[100*fileCountM1+j][1]+'\n')
      fp.write(self.data[100*fileCountM1+endNum-1][0]+'#'+self.data[100*fileCountM1+endNum-1][1])
      fp.close(); del fp, fileName
      indexFp.close(); del indexFp
    os.chdir(initCwd)


  def readOmnidic(self, filename, dicIndex=16):
    os.chdir(filename)
    try:
      fp = file(str(dicIndex))
    except:
      printAsError('bad index: %s'%dicIndex)
      return False
    for f in [l.split('#')[-1] for l in fp.read().split('\n')]:
      if f=='':
        continue
      for line in file(f).read().split('\n'):
        if line=='':
          pass
        elif line[0]=='#':
          pass
        else:
          parts = line.split('#')
          self.data.append((parts[0], string.join(parts[1:])))
    os.chdir(initCwd)


  def writePmd(self, filename='', writeInfo=True):
    if filename=='':
      filename = self.getInfo('name')+'.pmd'
      filename = filename.replace('/','_')
      filename = filename.replace('\\','_')
    if not os.path.isdir(filename):
      os.mkdir(filename)
    os.chdir(filename)
    maxLen=max([len(x[0]+x[1]) for x in self.data])
    indexFp = file('index', 'w')
    n = len(self.data)
    indexFp.write('%s\n'%n)
    s = 200
    (d,m)=divmod(n,s)
    if m>0:
      k2=d-1
    else:
      k2=d
    indexFp.write('%s\n'%275) #???????
    indexFp.write('%s\n'%(d-1))
    try:
      self.ui
    except AttributeError:
      self.ui = None
    ui = self.ui
    for k in range(k2):
      i = k*s
      indexFp.write('%s\t%s\n'%(self.data[i][0],i+1))
      fp = file(str(k), 'w')
      for j in range(i, i+s):
        fp.write('%s\x09%s\x00'%(self.data[j][0], pmdCompile(self.data[j][1].decode('utf-8'))))
      fp.close()
      if ui!=None:
          rat = float(k)/(k2+1)
          ui.progress(rat)
          while gtk.events_pending():
            gtk.main_iteration_do(False)
    indexFp.write(self.data[-1][0])
    indexFp.close()
    fp = file(str(d), 'w')
    for j in range(d*s, n):
      fp.write('%s\x09%s\x00'%(self.data[j][0], pmdCompile(self.data[j][1].decode('utf-8'))))
    fp.close()
    if ui!=None:    
      ui.progress(1.0, 'Converting Completed')
      while gtk.events_pending():
        gtk.main_iteration_do(False)
    os.chdir(initCwd)


  def writeDicformids(self, filename='', writeInfo=True):
    #'DictionaryForMIDs'
    indexFileMaxSize=30000
    language1FilePostfix='Eng'
    if filename=='':
      filename = self.getInfo('name')+'.pmd'
      filename = filename.replace('/','_')
      filename = filename.replace('\\','_')
    if not os.path.isdir(filename):
      os.mkdir(filename)
    os.chdir(filename)
    n=len(self.data)
    index=[None]*n
    dicMaxSize=0
    k = 1
    indFp = file('index%s1.csv'%language1FilePostfix, 'w')
    dicFp=file('directory%s1.csv'%language1FilePostfix, 'w')
    for i in range(n):
      if i%200==0 and i>0:
        #dicFp.close()
        dicSize=dicFp.tell()
        if dicSize > dicMaxSize:
          dicMaxSize = dicSize
        dicFp=file('directory%s%d.csv'%(language1FilePostfix, i/200), 'w')
      [w,m]=self.data[i][0:2]
      dicLine='%s\t%s\n'%(w,m)
      dicFp.write(dicLine)
      if i%200==0:
        index[i] = 0
      else:
        index[i] = index[i-1] + len(dicLine)
      indLine = '%s\t%d-%d-B\n'%(w,(i/200+1),index[i])
      size = indFp.tell()
      if size + len(indLine) > indexFileMaxSize-10 :
        k += 1
        #indFp.close()
        indFp = file('index%s%d.csv'%(language1FilePostfix, k) , 'w')
      indFp.write(indLine)
    dicFp.close()
    indFp.close()
    propFp=file('DictionaryForMIDs.properties', 'w')
    propStr='#DictionaryForMIDs property file\n' +\
            'infoText=%s, author: %s\n'%(self.getinfo('name'),self.getinfo('author'))+\
            'indexFileMaxSize=%s\n'%indexFileMaxSize +\
            'language1IndexNumberOfSourceEntries=%d\n'%n +\
            'language1DictionaryUpdateClassName=de.kugihan.dictionaryformids.dictgen.DictionaryUpdate\n'+\
            'indexCharEncoding=ISO-8859-1\n'+\
            "dictionaryFileSeparationCharacter='\\t'\n"+\
            'language2NormationClassName=de.kugihan.dictionaryformids.translation.Normation\n'+\
            'language2DictionaryUpdateClassName=de.kugihan.dictionaryformids.dictgen.DictionaryUpdate\n'+\
            'logLevel=0\n'+\
            'language1FilePostfix=%s\n'%language1FilePostfix +\
            'dictionaryCharEncoding=UTF-8\n'+\
            'numberOfAvailableLanguages=2\n'+\
            'language1IsSearchable=true\n' +\
            'language2GenerateIndex=false\n' +\
            'dictionaryFileMaxSize=%d\n'%(dicMaxSize+1) +\
            'language2FilePostfix=fa\n' +\
            'searchListFileMaxSize=20000\n' +\
            'language2IsSearchable=false\n' +\
            'fileEncodingFormat=plain_format1\n' +\
            'language1HasSeparateDictionaryFile=true\n' +\
            'searchListCharEncoding=ISO-8859-1\n' +\
            "searchListFileSeparationCharacter='\t'\n" +\
            "indexFileSeparationCharacter='\t'\n" +\
            'language1DisplayText=English\n' +\
            'language2HasSeparateDictionaryFile=false\n' +\
            'dictionaryGenerationInputCharEncoding=UTF-8\n' +\
            'language1GenerateIndex=true\n' +\
            'language2DisplayText=farsi\n' +\
            'language1NormationClassName=de.kugihan.dictionaryformids.translation.NormationEng\n' +\
    file('searchlist%s.csv'%language1FilePostfix, 'w') ### ??????????????????????????????????????
    os.chdir(initCwd)





  def readPmd(self, filename):
    os.chdir(filename)
    indLines = file('index').read().split('\n')[3:]
    n = len(indLines)-1
    for i in range(n):
      pmd = file(str(i)).read()
      for part in pmd.split('\x00'):
        sepInd = part.find('\x09')
        if sepInd==-1:
          printAsError('error on reading PMD database part %s'%i)
        word=part[:sepInd]
        mean=part[sepInd+1:]
        
        
    
     



########### This implemention may have trouble bugs, specially when converting from BGL
  def writeStardict_int(self, filename='', sort=True, dictZip=True):
    if sort:
      g = self.copy()
      g.data.sort()
    else:
      g = self
    if type(filename)!=str:
      raise TypeError, 'bad type given to method writeStardict: "%s"'%type(filename)
    if filename=='':
      filename = g.getInfo('name').replace('/','_').replace('\\','_')
    elif filename[-4:].lower()=='.ifo':
      filename=filename[:-4]
    elif filename[-1]==os.sep:
      if not os.path.isdir(filename):
        os.makedirs(filename)
      filename = filename + g.getInfo('name').replace('/','_').replace('\\','_')
    elif os.path.isdir(filename):
      filename = filename + os.sep + g.getInfo('name').replace('/','_').replace('\\','_')
    ifoFile =file(filename+'.ifo','w')
    idxFile =file(filename+'.idx','wb')
    dictFile=file(filename+'.dict','wb')
    dictMark=0
    idxStr=''
    dictStr=''
    for item in g.data:
        (word,mean)=item
        lm = len(mean)
        idxStr += ( word + '\x00' + intToBinStr(dictMark,4) + intToBinStr(lm,4) )
        dictMark += lm
        dictStr += mean
    dictFile.write(dictStr)
    dictFile.close()
    idxFile.write(idxStr)
    idxFile.close()
    ifoFile.write('StarDict\'s dict ifo file\nversion=2.4.2\n')
    ifoFile.write('wordcount=%d\n'%len(g.data))
    ifoFile.write('idxfilesize=%d\n'%len(idxStr))
    ifoFile.write('bookname=%s\n'%g.getInfo('name'))
    infoKeys = self.infoKeys()
    for key in Glossary.infoKeysStardict:
      value = g.getInfo(key)
      if value!='':
        ifoFile.write('%s=%s\n'%(key, value))
    ifoFile.write('sametypesequence=m\n')
    ifoFile.close()
    if dictZip:
      runDictzip(filename)

  def writeStardict(self, filename='', sort=True, dictZip=True):
    if sort:
      g = self.copy()
      g.data.sort()
    else:
      g = self
    if filename=='':
      filename=opt['filename']+'.ifo'
    try:
      import _stardictbuilder
    except ImportError:
      printAsError('Binary module "_stardictbuilder" can not be imported! '+\
        'Using internal StarDict builder')
      return g.writeStardict_int(filename, sort=False)
    db = _stardictbuilder.new_StarDictBuilder(filename)
    _stardictbuilder.StarDictBuilder_swigregister(db)
    for item in g.data:
      _stardictbuilder.StarDictBuilder_addHeadword(db,item[0],item[1],'')
    _stardictbuilder.StarDictBuilder_setTitle(db, g.getInfo('name'))
    _stardictbuilder.StarDictBuilder_setAuthor(db, g.getInfo('author'))
    _stardictbuilder.StarDictBuilder_setLicense(db, g.getInfo('license'))
    _stardictbuilder.StarDictBuilder_setOrigLang(db, '')
    _stardictbuilder.StarDictBuilder_setDestLang(db, '')
    _stardictbuilder.StarDictBuilder_setDescription(db, g.getInfo('description'))
    _stardictbuilder.StarDictBuilder_setComments(db, '')
    _stardictbuilder.StarDictBuilder_setEmail(db, '')
    _stardictbuilder.StarDictBuilder_setWebsite(db, '')
    _stardictbuilder.StarDictBuilder_setVersion(db, '')
    _stardictbuilder.StarDictBuilder_setCreationDate(db, '')
    _stardictbuilder.StarDictBuilder_setLastUpdate(db, '')
    _stardictbuilder.StarDictBuilder_finish(db)
    if dictZip:
      if filename[-4:]=='.ifo':
        filename = filename[:-4]
        runDictzip(filename)




  def readStardict(self, filename):
    self.data.sort()
    try:
      ifoStr =file(filename).read()
    except:
      try:
        ifoStr =file(filename+'.ifo').read()
      except:
        raise IOError, 'Can not read ifo file: %s'%filename
    else:
      filename = filename[:-4]
    try:
      idxStr = file(filename+'.idx', 'rb').read()
    except:
      raise IOError, 'Can not read idx file: %s'%filename
    try:
      dictStr=file(filename+'.dict', 'rb').read()
    except:
      import gzip
      dictStr=gzip.open(filename+'.dict.dz').read()
    for line in ifoStr.split('\n'):
      ind=line.find('=')
      if ind==-1:
        continue
      key=line[:ind]
      value=line[ind+1:]
      self.setInfo(key, value)
    self.data=[]
    word=''
    sumLen=0
    i=0
    errorsWords=[]
    ##############################  IMPORTANT PART ############################
    while i < len(idxStr):
      if idxStr[i]=='\x00':
        sumLen2 = binStrToInt(idxStr[i+1:i+5])
        meanLen=binStrToInt(idxStr[i+5:i+9])
        if sumLen2 != sumLen:
          printAsError('Error on word="%s" i=%d sumLen=%d sumLen2=%d len(self.data)=%d'\
            %(word,i,sumLen,sumLen2,len(self.data)))
          errorsWords.append(word)
          mean=''
        else:
          mean=dictStr[sumLen:sumLen+meanLen].replace('<BR>','\n')
        self.data.append((word,mean))
        sumLen = sumLen2
        sumLen += meanLen
        word = ''
        i += 9
      else:
        word += idxStr[i]
        i += 1
    ############################################################################
    if len(errorsWords)>0:
      print 'errorsCount=%d'%len(errorsWords)
      print 'errorsWords=%s'%errorsWords
    if filename[-4:].lower()=='.ifo':
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)




  def readStardict_ext(self, filename):
    ## This method uses module provided by dictconv, but dictconv is very slow for reading from stardict db!
    ## therefore this method in not used by GUI now.
    ## and binary module "_stardict" is deleted.
    ## If you want to use it, compile it yourself, or get it from an older version of PyGlossary (version 2008.08.30)
    self.data.sort()
    import _stardict
    db = _stardict.new_StarDict(filename)
    _stardict.PySwigIterator_swigregister(db)
    self.setInfo('title', _stardict.StarDict_bookname(db))
    self.setInfo('author', _stardict.StarDict_bookname(db))
    self.setInfo('title', _stardict.StarDict_bookname(db))
    ## geting words:
    idxStr = file(filename[:-4]+'.idx').read()
    words=[]
    #""" # get words list by python codes.
    word=''
    i=0
    while i < len(idxStr):
      if idxStr[i]=='\x00':
        if word!='':
          words.append(word)
          word = ''
        i += 9
      else:
        word += idxStr[i]
        i += 1
    """ # get words list using _stardict module.
    lst = _stardict.StarDict_dump(db) ## don't know how to get std::vector items
    _stardict.PySwigIterator_swigregister(lst)
    num = _stardict.StarDict_vector_get_size(db, lst)
    _stardict.PySwigIterator_swigregister(num)
    for i in range(num):
      word = _stardict.StarDict_vector_get_item(db, lst, i)
      words.append(word)
      if i%100==0:
        print i
        while gtk.events_pending():
          gtk.main_iteration_do(False)
    """
    i=0
    for word in words:
      mean = _stardict.StarDict_search(db, word).replace('<BR>','\n')
      self.data.append((word,mean))
    if filename[-4:].lower()=='.ifo':
      filename = filename[:-4]
    self.setInfo('filename', filename)
    if self.getInfo('name') == '':
      name = os.path.split(filename)[-1]
      self.setInfo('name', name)


  def writeMdic_ext(self, filename=''):
    ## This method uses binary module "_mdicbuilder"
    ## but this binary module is deleted from package PyGlossary, and not used by GUI now.
    ## If you want to use it, compile it yourself, or get it from an older version of PyGlossary (version 2008.08.30)
    if filename=='':
      filename=self.getInfo('filename')+'.m2'
    import _mdicbuilder
    from _mdicbuilder import MDicBuilder_addHeadword
    if os.path.exists(filename):
      os.remove(filename)
    db = _mdicbuilder.new_MDicBuilder(filename)
    _mdicbuilder.MDicBuilder_swigregister(db)
    n = len(self.data)
    try:
      self.ui
    except AttributeError:
      self.ui = None
    ui = self.ui
    if ui==None:
      for i in range(n):
        MDicBuilder_addHeadword(db,self.data[i][0],self.data[i][1].replace('\n','<BR>'),'')
    else:
      k = 1000
      for i in range(n):
        MDicBuilder_addHeadword(db,self.data[i][0],self.data[i][1].replace('\n','<BR>'),'')
        if i%k==0:
          rat = float(i)/n
          ui.progress(rat)
          while gtk.events_pending():
            gtk.main_iteration_do(False)
      ui.progress(1.0, 'Converting Completed')
      while gtk.events_pending():
        gtk.main_iteration_do(False)
    _mdicbuilder.MDicBuilder_setTitle(db, self.getInfo('name'))
    _mdicbuilder.MDicBuilder_setAuthor(db, self.getInfo('author'))
    _mdicbuilder.MDicBuilder_setLicense(db, self.getInfo('license'))
    _mdicbuilder.MDicBuilder_setOrigLang(db, '')
    _mdicbuilder.MDicBuilder_setDestLang(db, '')
    _mdicbuilder.MDicBuilder_setDescription(db, self.getInfo('description'))
    _mdicbuilder.MDicBuilder_setComments(db, '')
    _mdicbuilder.MDicBuilder_setEmail(db, '')
    _mdicbuilder.MDicBuilder_setWebsite(db, '')
    _mdicbuilder.MDicBuilder_setVersion(db, '')
    _mdicbuilder.MDicBuilder_setCreationDate(db, '')
    _mdicbuilder.MDicBuilder_setLastUpdate(db, '')
    _mdicbuilder.MDicBuilder_finish(db)



  def writeMdic(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.m2'
    from sqlite3 import connect
    keys=('name', 'author', 'version', 'direction', 'origLang', 'destLang', 'license', 'category', 'description')
    if os.path.exists(filename):
      os.remove(filename)
    con = connect(filename)
    cur = con.cursor()
    sqlLines = self.writeSql(None).split('\n')
    n = len(sqlLines)
    try:
      self.ui
    except AttributeError:
      self.ui = None
    ui = self.ui
    if ui==None:
      for i in range(n):
        #try:
        cur.execute(sqlLines[i][:-1])
        #except:
        #  print sqlLines[i]
        #  return False
    else:
      k = 1000
      for i in range(n):
        con.execute(sqlLines[i][:-1])
        if i%k==0:
          rat = float(i)/n
          ui.progress(rat)
          while gtk.events_pending():
            gtk.main_iteration_do(False)
      ui.progress(1.0, 'Converting Completed')
      while gtk.events_pending():
        gtk.main_iteration_do(False)
    cur.close()
    con.close()



  def readMdic(self, filename):
    from sqlite3 import connect
    keys=('name', 'author', 'version', 'direction', 'origLang', 'destLang', 'license', 'category', 'description')
    con = connect(filename)
    cur = con.cursor()
    for key in keys:
      try:
        cur.execute('select %s from dbinfo'%key)
      except:
        pass
      else:
        value = cur.fetchone()[0].encode('utf8')
        if value!='':
          self.setInfo(key, value)
    cur.execute('select * from word')
    self.data = [(x[1].encode('utf8'), x[2].encode('utf8')) for x in cur.fetchall()]
    cur.close()
    con.close()

  def writeSql(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.sql'
    keys=('name', 'author', 'version', 'direction', 'origLang', 'destLang', 'license', 'category', 'description')
    sqlStr=''
    #sqlStr+='BEGIN TRANSACTION;\n'
    #sqlStr+='CREATE TABLE dbinfo ('
    #for key in keys[:-1]:
    #  sqlStr += ('%s varchar(256), '%key)
    #sqlStr += '%s varchar(256));\n'
    sqlStr += ('CREATE TABLE dbinfo (dbname varchar(50), author varchar(50), version varchar(8), direction char(3), '+\
      'origLang char(5), destLang char(5), license varchar(50), category varchar(50), description varchar(200));\n')
    sqlStr+="CREATE TABLE word ('s_id' INTEGER PRIMARY KEY NOT NULL, 'wname' TEXT, 'wmean' TEXT);\n"
    sqlStr += ('INSERT INTO dbinfo VALUES(')
    for key in keys[:-1]:
      sqlStr += ("'%s',"%self.getInfo(key))
    sqlStr+=("'%s');\n"%self.getInfo(keys[-1]))
    for i in range(len(self.data)):
      x = self.data[i]
      sqlStr += ("INSERT INTO word VALUES(%d,'%s','%s');\n"%(i+1, x[0].replace('\'','\'\''), x[1].replace('\'','\'\'')))
    sqlStr += ("CREATE INDEX wnameidx ON word(wname);")
    #sqlStr += "COMMIT;"
    if filename==None:
      return sqlStr
    fp=file(filename, 'w')
    fp.write(sqlStr)
    fp.close()
    return True


  def writeIndex(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.index'
    fp=file(filename, 'w')
    sumLen=0
    for item in self.data:
      l = len(item[1])
      fp.write(item[0]+'\t'+intToIndexStr(sumLen)+'\t'+intToIndexStr(l)+'\n')
      sumLen += l
    fp.close()

  def writeTreedict(self, filename='', archive='tar.bz2', sep=os.sep):
    if filename=='':
      filename=self.getInfo('filename')+'.treedict'
    if os.path.exists(filename):
      if os.path.isdir(filename):
        if os.listdir(filename)!=[]:
          print 'Warning: directory "%s" is not empty.'
      else:
        raise IOError, '"%s" is not a directory'
    for item in self.data:
      if item[0]=='':
        printAsError('empty word')
        continue
      chars = list(item[0])
      try:
        os.makedirs(filename+os.sep+string.join(chars[:-1], sep))
      except:
        pass
      try:
        file('%s%s%s.m'%(filename,os.sep,string.join(chars, sep)), 'w').write(item[1])
      except:
        print item[1]
    if not archive in (None, ''):
      if archive=='tar.gz':
        os.popen('tar -czf \'%s.tar.gz\' \'%s\''%(filename,filename))
      elif archive=='tar.bz2':
        os.popen('tar -cjf \'%s.tar.bz2\' \'%s\''%(filename,filename))
      elif archive=='zip':
        os.popen('zip -r \'%s.zip\' \'%s\''%(filename,filename))
      else:
        printAsError('Undefined archive format: "%s"'%archive)
      try:
        os.popen("rm -R 'filename'")
      except:
        pass

  def writeFreedict(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.tei'
    fp=file(filename, 'w')
    fp.write('<?xml version="1.0" encoding="UTF-8"?>\n'+
    '<!DOCTYPE TEI.2 PUBLIC "-//TEI P3//DTD Main Document Type//EN"\n'
    '"/usr/share/sgml/tei-3/tei2.dtd" [\n<!ENTITY % TEI.dictionaries "INCLUDE" > ]>\n'+
    '<tei.2>\n<teiHeader>\n<fileDesc>\n  <titleStmt>\n    <title>%s</title>\n'%self.getInfo('title')+
    '    <respStmt><resp>converted with</resp><name>PyGlossary</name></respStmt>\n  </titleStmt>\n'+
    '  <publicationStmt><p>freedict.de</p></publicationStmt>\n'+
    '  <sourceDesc><p>%s</p></sourceDesc>\n</fileDesc>\n</teiHeader>\n<text><body>'%filename)
    for item in self.data:
      fp.write('<entry>\n  <form><orth>%s</orth></form>\n  <gramgrp><pos>n</pos></gramgrp>\n  <trans><tr>%s</tr></trans>\n</entry>'%item[0:2])
    fp.write('</body></text></tei.2>')
    fp.close()

  def writeGettextPo(self, filename=''):
    if filename=='':
      filename=self.getInfo('filename')+'.tei'
    fp=file(filename, 'w')
    fp.write('#\nmsgid ""\nmsgstr ""\n')
    for inf in self.infoKeys():
      fp.write('"%s: %s\\n"'%(inf, self.getInfo(inf)))
    for item in self.data:
      fp.write('msgid "%s"\nmsgstr "%s"\n\n'%item[:2])
    fp.close()

  def readGettextPo(self, filename):
    pass


