//   Encoding: UTF-8
//   File: reverse_dic.cpp 
//
//   Author: Saeed Rasooli <saeed.gnu@gmail.com>  (ilius)
//
//   This program is a free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 3, or (at your option)
//   any later version.
//
//   You can get a copy of GNU General Public License along this program
//   But you can always get it from http://www.gnu.org/licenses/gpl.txt
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  the
//   GNU General Public License for more details.

#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include "reverse_dic.h"

using namespace std;

string int2str(int n){
  ostringstream oss;
  oss << n;
  return oss.str();
}
int str2int(string s){
  int n;
  stringstream ss;
  ss << s;
  ss >> n;
  return n;
}
int str2int(char *s){
  int n;
  stringstream ss;
  ss << s;
  ss >> n;
  return n;
}
int float2int(float f){
  int n;
  stringstream ss;
  ss << f;
  ss >> n;
  return n;
}
string float2str(float f){
  ostringstream oss;
  oss << f;
  return oss.str();
}


vector<string> split(string str, string sep=" ", bool empty=true){
  vector<string> v;
  int i, i2, n, ss=sep.size();
  string last;
  n=str.size();
  v.clear();
  i=0;
  while(true){
    i2=str.find(sep, i);
    //cout << i << "\t" << i2 << endl;
    if(i2==-1){
      if(i<n)
        v.push_back(str.substr(i, n-i));
      break;
    }
    v.push_back(str.substr(i, i2-i));
    i=i2+ss;
  }
  //if(str.substr(n-1,1)==sep)
  //  v.push_back("");
  return v;
}

string replace(string str, string sstr, string tstr){
  int i=0, i2, ss=sstr.size();
  while(true){
    i2=str.find(sstr, i);
    //cout << i << "\t" << i2 << endl;
    if(i2==-1)
      break;
    str.replace(i2, ss, tstr);
    i=i2+ss;
  }
  return str;
}


int findCount(string str, string sstr){
  int i=0, i2, ss=sstr.size(), n=0;
  while(true){
    i2=str.find(sstr, i);
    //cout << i << "\t" << i2 << endl;
    if(i2==-1)
      break;
    n++;
    i=i2+ss;
  }
  return n;
}


glossary readTabStr(string tstr){
  int i, n, pos, ll;
  string line;
  vector<string> lines;
  //entry e;
  glossary g;
  lines = split(tstr, "\n");
  n = lines.size();
  cout << "Tabfile lines count: " << n << endl;
  for(i=0;i<n;i++){
    line = lines[i];
    ll = line.size();
    pos = line.find("\t");
    entry e;
    e.word = line.substr(0, pos);
    e.definition = line.substr(pos, ll-pos);
    g.data.push_back(e);
  }
  return g;
}




string Reverse::Reverse(string a_tabStr, float a_minRel=0.1, int a_maxNum=100, string a_sep="،", bool a_matchWord=true, string a_showRel="Percent", string a_fifoPath=""){
  tabStr=a_tabStr;
  minRel=a_minRel;
  maxNum=a_maxNum;
  sep=a_sep;
  matchWord=a_matchWord;
  fifoPath=a_fifoPath;
}


string Reverse::~Reverse(){}

string Reverse::search(string st, bool multiWord=true){
  int i, j, k, n, size, pNum, pLen, stSize=st.size(), index, iRel, piRel;
  entry e;
  string word, mean, part, part2, outStr("");
  glossary g=readTabStr(tabStr);
  vector<string> meanParts, partWords, relWords;
  float rel, partRel, prevRel;
  n = g.data.size();
  meanParts.clear();
  partWords.clear();
  relWords.clear();
  char relCh[2];
  cout << "g.data size: " << n << endl;
  for(i=0;i<n;i++){
    //cout << "i=" << i << endl;
    word = g.data[i].word;
    mean = g.data[i].definition;
    meanParts = split(mean, sep);
    if(mean.find(st)==-1)
      continue;
    rel=0.0;
    size=meanParts.size();
    for(j=0;j<size;j++){
      part = meanParts[j];
      for(k=0; k<60; k++)
        part=replace(part, sch[k], " ");
      partRel=0.0;
      if(matchWord){
        pNum=0;
        partWords=split(part, " ");
        pLen=partWords.size();
        if(pLen==0)
          continue;
        for(k=0;k<partWords.size();k++)
          if(partWords[k]==st)
            pNum++;
        partRel = ((float)pNum)/pLen;
      }
      else{
        part2.assign(part);
        part=replace(part2, " ", "");
        pLen=part2.size();
        if(pLen==0)
          continue;
        pNum = findCount(part, st)*stSize;
        partRel = ((float)pNum)/pLen;
      }
      if(partRel>rel)
        rel=partRel;
    }
    if(rel<minRel)
      continue;
    if(rel==1.0)
      rel=0.99;
    part = int2str( 99-float2int(rel*100.0 )) + " " + word ;
    //cout << rel << "\t" << part << endl;
    relWords.push_back(part);
    meanParts.clear();
    partWords.clear();
  }
  sort(relWords.begin(), relWords.end());
  n = relWords.size();
  cout << "relWords size: " << n << endl;
  if(maxNum>0 && maxNum<n){
    for(j=maxNum; j<n; j++)
      relWords.pop_back();
    n=maxNum;
  }
  iRel=0;
  outStr="";
  sep=", ";
  for(j=0;j<n;j++){
    part = relWords[j];
    if(j==n-1)
      sep=".";
    piRel = rel;
    iRel = 99 - str2int(part.substr(0, 2));
    word=part.substr(3, part.size()-3);
    if(iRel==99)
      outStr += (word+sep);
    else if(showRel=="Percent")
      outStr += (word + "(%" + int2str(iRel) + ")" + sep);
    else if(showRel=="Percent At First"){
      if(iRel==piRel)
        outStr += (word+sep);
      else
        outStr += (word + "(%" + int2str(iRel) + ")" + sep);
    }
    else
      outStr += (word+sep);
  }
  return outStr;
}


/*
int main(int argc, char **argv){
  int i, bufSize=5000000, k;
  string text, st;
  vector<string> v;
  FILE *fp;  
  char buf[bufSize];
  fp=fopen("/media/MyFlash8/dic-work/tmp-dic/quick_eng-persian.txt", "r");
  fread(buf, 1, bufSize, fp);
  text=(string)buf;
  cout << "file text size: " << text.size() << endl;
  cout << search(text, (string)"سلام") << endl;
  //cout << str2int(argv[1])*10 << endl;
  //cout << replace("hello world", "l", "L") << endl;
  
  return 0;
}
*/

